#!/usr/bin/env python
# encoding: utf-8

import numpy as np
import _edflib as edf
from _edflib import FILETYPE_EDFPLUS
import pprint


class EDFReader(object):
    def __init__(self, file_name):
        self.edf = edf.Edfreader(file_name)

    def get_signal_info(self):
        return {i: {"samples": self.edf.samples_in_file(i),
                    "sample_rate": self.edf.samplefrequency(i),
                    "label": self.edf.signal_label(i)}
                for i in range(self.edf.signals_in_file)}

    def file_info(self):
        pprint.pprint(self.get_signal_info())
        pprint.pprint(self.get_meta_data())

    def read_signal(self):
        signal_info = self.get_signal_info()
        signal_buffers = {signal_number: np.empty(info["samples"]) for signal_number, info in signal_info.items()}

        for signal_number, signal_buffer in signal_buffers.items():
            samples = signal_info[signal_number]["samples"]
            self.edf.readsignal(signal_number, 0, samples, signal_buffer)

        return signal_buffers

    def get_meta_data(self):
        return {"patient_name": self.edf.patientname,
                "equipment": self.edf.equipment,
                "patientcode": self.edf.patientcode,
                "gender": self.edf.gender,  # 1: male, 0: female
                "patient_birthdate": self.edf.birthdate,
                "recording_additional": self.edf.recording_additional,
                "patient_additional": self.edf.patient_additional,
                "admincode": self.edf.admincode,
                "technician": self.edf.technician,
                "duration": self.edf.file_duration}

    def get_annotations(self):
        return self.edf.annotations_in_file


class EdfWriter(object):
    def __init__(self, file_name, channel_info, meta_data=None, annotations=None, file_type=edf.FILETYPE_EDFPLUS):
        """Initialises an EDF file at @file_name.
        @file_type is one of
            edflib.FILETYPE_EDF
            edflib.FILETYPE_EDFPLUS
            edflib.FILETYPE_BDF
            edflib.FILETYPE_BDFPLUS

        @channel_info should be a
        list of dicts, one for each channel in the data. Each dict needs
        these values:

            'label' : channel label (string, <= 16 characters, must be unique)
            'dimension' : physical dimension (e.g., mV) (string, <= 8 characters)
            'sample_rate' : sample frequency in hertz (int)
            'physical_max' : maximum physical value (float)
            'physical_min' : minimum physical value (float)
            'digital_max' : maximum digital value (int, -2**15 <= x < 2**15)
            'digital_min' : minimum digital value (int, -2**15 <= x < 2**15)
        """
        self.file_type = file_type
        self.n_channels = len(channel_info)
        self.channels = {}
        for c in channel_info:
            self.channels[c['label']] = c
        self.sample_buffer = dict([(c['label'], np.array([])) for c in channel_info])
        self.handle = edf.open_file_writeonly(file_name, file_type, self.n_channels)
        if self.handle < 0:
            raise IOError("edf file could not be opened: %s" % file_name)
        self._init_channels(channel_info)
        self._init_annotations(annotations)
        self.active = True
        if meta_data:
            self._init_constants(**meta_data)

    def _init_annotations(self, annotations):
        # see http://www.teuniz.net/edflib/#edfwrite_annotation_latin1
        if annotations:
            for annotation in annotations:
                edf.write_annotation_latin1(self.handle,
                                            annotation.onset * 10000,
                                            -1 if annotation.duration == -1 else annotation.duration * 10000,
                                            annotation.text + "\0")

    def write_samples(self, channel_label, samples):
        """Queues a digital sample for @channel_label for recording; the data won't
        actually be written until one second's worth of data has been queued."""
        if self.active:
            self.sample_buffer[channel_label] = np.concatenate((self.sample_buffer[channel_label], samples))
            self.check_samples()

    def check_samples(self):
        channels_full = [len(self.sample_buffer[c]) / int(self.channels[c]['sample_rate']) for c in self.channels]
        if min(channels_full) >= 1:
            self._flush_samples(seconds=min(channels_full))

    def close(self):
        self.active = False
        # make sure that last, potentially incomplete second is also written
        # ensure this by filling up the buffer to one second of data
        self.check_samples()
        sorted_channel_keys = sorted(self.channels.keys(), key=lambda x: int(x))
        for c in sorted_channel_keys:
            sample_rate = self.channels[c]["sample_rate"]
            padding = sample_rate - len(self.sample_buffer[c])
            if len(self.sample_buffer[c]) and padding > 0:
                fill_array = np.empty(padding)
                fill_array.fill(self.sample_buffer[c][-1])
                self.sample_buffer[c] = np.concatenate((self.sample_buffer[c], fill_array))
        self.check_samples()
        if not all([len(self.sample_buffer[c]) == 0 for c in self.channels]):
            print "last second of edf data was not written completely"
            print "sampling rate", [self.channels[c]["sample_rate"] for c in self.channels]
            print "sample count", [len(self.sample_buffer[c]) for c in self.channels]
            print "number of lost seconds", [len(self.sample_buffer[c]) / float(self.channels[c]["sample_rate"]) for c in self.channels]
        edf.close_file(self.handle)

    def _flush_samples(self, seconds=1):
        for _ in xrange(seconds):
            sorted_channel_keys = sorted(self.channels.keys(), key=lambda x: int(x))
            for c in sorted_channel_keys:
                sample_rate = self.channels[c]["sample_rate"]
                buf = np.array(self.sample_buffer[c][:sample_rate], dtype='int32')
                edf.write_digital_samples(self.handle, buf)
                self.sample_buffer[c] = self.sample_buffer[c][sample_rate:]

    def _init_constants(self, **kwargs):
        def call_if_set(fn, kw_name):
            item = kwargs.pop(kw_name, None)
            if item is not None:
                if kw_name == "gender" and item not in [0, 1]:
                    return
                fn(self.handle, item)
        call_if_set(edf.set_technician, 'technician')
        call_if_set(edf.set_recording_additional, 'recording_additional')
        call_if_set(edf.set_patientname, 'patient_name')
        call_if_set(edf.set_patientcode, 'patientcode')
        call_if_set(edf.set_patient_additional, 'patient_additional')
        call_if_set(edf.set_equipment, 'equipment')
        call_if_set(edf.set_admincode, 'admincode')
        call_if_set(edf.set_gender, 'gender')
        call_if_set(edf.set_datarecord_duration, 'duration')
        call_if_set((lambda hdl, dt: edf.set_startdatetime(hdl, dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)), 'recording_start_time')
        call_if_set((lambda hdl, dt: edf.set_birthdate(hdl, dt.year, dt.month, dt.day)), 'patient_birthdate')
        if len(kwargs) > 0:
            raise Exception('Unhandled argument(s) given: %r' % kwargs.keys())

    def _init_channels(self, channels):
        hdl = self.handle

        def call_per_channel(fn, name, optional=False):
            for i, c in enumerate(channels):
                if optional and not name in c:
                    continue
                fn(hdl, i, c[name])

        call_per_channel(edf.set_samplefrequency, 'sample_rate')
        call_per_channel(edf.set_physical_maximum, 'physical_max')
        call_per_channel(edf.set_digital_maximum, 'digital_max')
        call_per_channel(edf.set_digital_minimum, 'digital_min')
        call_per_channel(edf.set_physical_minimum, 'physical_min')
        call_per_channel(edf.set_label, 'label')
        call_per_channel(edf.set_physical_dimension, 'dimension')
        call_per_channel(edf.set_transducer, 'transducer', optional=True)
        call_per_channel(edf.set_prefilter, 'prefilter', optional=True)
